% lab12_ex_adapt_demo.m
%
% ADAPTIVE FILTERS - QUICK INTRODUCTION
%
% SUMMARY
%
% The typical FIR digital filtration can be interpreted as a local weighted averaging
% some last previous input samples x():
%            M
%    y(n) = SUM( h(m)*x(n-m) ),   y(n) = h0*x(n-0) + h1*x(n-1) + ... + hM*x(n-M))
%           m=0
% In adaptive FIR filters weights h(m), m=0...M, are specially designed (for: LP,HP, BP,
% BS filters, Hilbert phase shifter, digital differentiator) and ARE CONSTANT during
% usage.
% In ADAPTIVE filters the weights h(m), m=0...M, are changing from sample to sample and
% can be different for different time index n.
%
% Adaptive filter has two inputs:
%     x(n) - signal to be filtered,
%     d(n) - referenced (desired) signal, not processed,
% and two outputs:
%     y(n) - filtered x(d),
%     e(n) = difference between d(n) and y(n): e(n) = d(n) - y(n).
%  
%     d(n) --> [ ------->(+)---> ] --> e(n)
%              [        + ^-     ]
%              [          |      ]
%     x(n) --> [ --> H(z) |----> ] --> y(n)
%
% The filter weights h(m) adaptation:
%     h(m,n+1) = h(m,n) + dh(m,n),  new(n+1) = old(n) + change(n),
% should ensure minimization of the chosen cost error function J(n):
% - least mean squared error J(n) = E{e(n)^2},
% - cumulative least squared error from the filter start:
%     J(n) = SUM{k=0...n}(e(k)^2)
% - cumulative weighed least squared error, forgetting the past:
%     J(n) = SUM{k=0...n}(lambda^(n-k)*e(k)^2)
%
% Therefore there are the following adaptive filters:
% (N)LMS - (normalized) least mean squares filters, with slow convergence, a few +-*/;
%    WLS - weighted least squares filters, with faster convergence but with more +-*/;
%   RWLS - recursive WLS, with smaller number of +-*/ in comparison to WLS.
%
% LMS ADAPTIVE FILTERS
%
% In case of the LMS filters it is assumed the minimization of the squared instantaneous
% error value e(n) should lead to the minimization of the mean squared error value
% (therefore operator E{.} is removed):
%
%   J(n) = e(n)^2 = ( d(n) - h0*x(n-0) - h1*x(n-1) - ... - hM*x(n-M) )^2
%
% Since we are looking for a minimum of a non-negative convex function J(n),
% it is sufficient to calculate its partial derivatives in respect to all coefficients
% h(k), k=0...M, i.e. gradients - directions of function growth,
% and to go in the opposite direction (minus sign) during adaptation of filter coeffs:
%
%   dJ(n)/d_hk = 2*e(n) * diff( d(n)-h0*x(n-0)-h1*x(n-1)-...-hM*x(n-M) )/diff(hk),
%   dJ(n)/d_hk = -2*e(n)*x(n-k),
%   hk(n+1)    = hk(n) - dJ(n)/d_hk,
%   hk(n+1)    = hk(n) + 2*mi*e(n)*x(n-k)
%
% where mi denotes speed adaptation factor.
% Fulfillment of the following convergence condition is required:
%
%   0 < mi < 2/lambda_max,  0 < mi < 2/trace(Rxx),
%
% where lambda_max is the maximum eigenvalue of the autocorrelation matrix Rxx
% of signal x(n).
% Larger mi value offers faster adaptation convergence but is more risky from
% the filter stability point of view and makes the filter more sensitive to noise
% (the filter is fast and attempts to track the noise, offering noisy outputs).
% It is important to note that speed of convergence of adaptive filter is proportional
% to the ratio lambda_max/lambda_min. What tells that the convergence is faster for
% x(n) being a noisy (random) signal.
%
% For stationary signals one can directly calculate optimal weights h_opt
% to which the filter should converge. It can be done by:
%
% 1) re-writing the cost function into the matrix form: 
%    J(h) = E{d(n)^2} - 2*h^T*rdx + h^T*Rxx*h,
%    where:
%    h   - vector of filter weights
%    rdx - vector of cross correlation between d(n) and x(n),
%    Rxx - matrix of autocorrelation of x(n),
%
% 2) calculation J(h) derivative in respect to vector h and setting it to zero:
%    dJ(h))/dh = -2*rdx + 2*Rxx*h = 0,
%
% 3) solving the last equation is respect to h (so-called optimal Wiener filter):
%    h_opt = inv(Rxx) * rdx = Rinv * rdx.
%
% (R)LS ADAPTIVE FILTERS
%
% In LS algorithms filter weights h are estimated using the last equation:
%    h(n) = Rinv(n) * rdx(n).
% and estimates of Rinv and rdx valid for time (n).
% To make the algorithm run faster the Rinv(n+1) and rdx(n+1) are calculated as
% recursive modification of Rinv(n) and rdx(n):
%    Rinv(n+1) = Rinv(n) + Rcorrection( x(n+1) ),
%    rdx(n+1)  = rdx(n)  + rcorrection( x(n+1) ),
% and adaptation of h takes place:
%    h(n+1) = Rinv(n+1) * rdx(n+1).
% The last equation can be re-written into the following form:
%    h(n+1) = h(n) + Rinv(n+1) * x(n+1) * e(n+1).
% Estimates of Rinv(n+1) and rdx(n+1) can be forgetting their past values with speed
% depending on the weighting parameter 0<=lambda<=1.
% Such algorithm is called Weighted Recursive Least Squares (WRLS).
% When lambda=1, the WRLS filters gives the best LS estimate of h(n), based on the
% observation time interval from 0 to n.
% The smaller lambda is, the WRLS filters is more forgetting the past.
% In such situation it more dynamic and faster adapting but for the price of bigger
% dependence to noise (bigger variance in stationary states).
%
% APPLICATIONS
%
% Adaptive interference (e.g.echo) cancelling.
% Adaptive signal enhancement.

  clear all; close all;
  
% Run program for all combinations of parameters ialg  and  itest
  itask = 1;        % task: 1 - adaptive interference cancelling (power supply cross-talk)
                    %       2 - adaptive signal de-noising using linear prediction
  ialg = 1;         % adaptation algorithm:
                    %       1 = LMS,
                    %       2 = NLMS (normalized)
                    %       3 = RLS

% LMS filter (Least Mean Squares)
  M = 50;           % number of filter coefficients (weights)
  mi = 1;        % adaptation speed coefficient ( 0<mi<1)

% NLMS filter (normalized LMS)
  gamma = 0.001;    % security constant e.g. = 0.001 (ensuring that denominator != 0)

% RLS filter (recursive LS) - more-computational demending but faster converging
  lambd = 0.999;        % RLS - forgetting factor for Least-Squares cost function 
  Rinv = 0.5*eye(M,M);  % RLS - inverse of Rxx matrix
  
% Generation of an LFM (linear frequency modulation) test signal
  Nx = 1000;        % number of samples
  fs = 1000;        % sampling frequency
  A = 1;            % amplitude
  f0 = 0;           % initial signal frequency
  df = 100;         % frequency increase per second [Hz/s]
  env  = 1;         % choose signal envelope: 0=rectang, 1=Gauss, 2=alfa*t, 3=exp(-alfa*t)
  
  f=0:fs/500:fs/2;                                       % frequencies for ploting
  dt=1/fs; t=0:dt:(Nx-1)*dt;                             % time points
  s = A*cos( 2*pi* (f0*t + 0.5*df*t.^2) );               % SIGNAL
  
  if (env == 0) w = boxcar(Nx)'; end                     % ENVELOPE 0 - rectangular
  if (env == 1) alfa=10; w=exp(-alfa*pi*(t-0.5).^2); end % envelope 1 - Gauss
  if (env == 2) alfa=5; w=alfa*t; end                    % envelope 2 - alfa*t
  if (env == 3) alfa=5; w=exp(-alfa*t); end              % envelope 3 - exp(-alfa*t)
  
  s = s .* w;                       % SIGNAL WITH ENVELOPE
  if (itask==1)                     % TEST 1 - interference cancelling
     P = 0;                         % no prediction
     x = 0.1*sin(2*pi*200*t-pi/5);  % power supply shifted in phase and scaled in amplitude
     d = s + 0.5*sin(2*pi*200*t);   % signal + interference
  end
  if (itask==2)                     % TEST 2 - de-noising by prediction
     P = 1;                         % prediction order set to 1,2,3,...)
     x = s + 0.25*randn(1,Nx);      % signal + noise
     d = [ x(1+P:length(x)) 0 ];    % signal x shifted in time by P samples (earlier)
  end
  
% Figures  input signals
  figure;
  subplot(211); plot(t,x); grid; title('IN : signal x(n)');
  subplot(212); plot(t,d); grid; title('IN : signal d(n)');
  xlabel('time [s]'); pause

% Calculation of optimal Wiener filter and limits for mi
  for k = 0 : M
   rxx(k+1) = sum( x(1+k:Nx) .* x(1:Nx-k) )/(Nx-M);  % auto-correlation of x(n)
   rdx(k+1) = sum( d(1+k:Nx) .* x(1:Nx-k) )/(Nx-M);  % cross-correlation of d(n) and x(n)
  end
  Rxx = toeplitz(rxx,rxx);           % symetrical autocorrelation matrix of x(n)
  h_opt = Rxx\rdx';                  % weights of the optimal Wiener filter
  lambda = eig( Rxx );               % eigenvalue decomposition of Rxx
  lambda = sort(lambda,'descend');   % sorting eigenvalues
  disp('Suggested values of mi')
  mi1_risc = 1/lambda(1),
  mi2_risc = 1/sum(lambda), pause
  figure;
  subplot(211); stem( h_opt ); grid; title('Optimal Wiener filter h(n)');
  subplot(212); stem( lambda ); grid; title('Eigenvalues of matrix Rxx');
% mi = mi2_risc/20;
  
% Adaptive filtration
  bx = zeros(1,M);  %   initialization of buffer for input signal x(n)
  h  = zeros(1,M);  %   initialization of filter weights (coefficients)
  y = [];
  e = [];
  for i = 1 : length(x)           % Main loop
      bx = [ x(i) bx(1:M-1) ];    % put new sample of x(n) into the buffer
      dest = h * bx';             % filtration of x(n) = prediction of d(n)
      err = d(i) - dest;          % prediction error
      if (ialg==1)  % LMS 
          h = h + ( 2*mi * err * bx );               % LMS  - weights adaptation
      end      
      if (ialg==2)  % NLMS
         eng = bx * bx';                             % NLMS - signal energy in bx 
         h = h + ( (2*mi)/(gamma+eng) * err * bx );  % NLMS - weights adaptation
      end
      if (ialg==3)  % RLS
          Rinv = (Rinv - Rinv*bx'*bx*Rinv/(lambd+bx*Rinv*bx'))/lambd; % modify Rinv 
	      h = h + (Rinv * bx' * err )';	                              % modify eights
    end    
    % subplot(211); stem(h); xlabel('n'); title('h(n)'); grid;
    % subplot(212); plot(f,abs(freqz(h,1,f,fs))); xlabel(' f [Hz]'); title('|H(f)|'); grid; pause
      y = [y dest];
      e = [e err];
  end
  
% Figures  output signals
  figure;
  subplot(211); plot(t,y); grid; title('OUT : signal y(n) = dest');
  subplot(212); plot(t,e); grid; title('OUT : signal e(n) = err');
  xlabel('time [s]'); pause
  if (itask==1)
     figure; subplot(111); plot(t,s,'r',t,e,'b'); grid; xlabel('time [s]');
     title('Original (RED), filtration result (BLUE)'); pause
  end
  if (itask==2)
     n=Nx/2+1:Nx; 
     SNR_we_dB = 10*log10( sum( s(n).^2 ) / sum( (d(n)-s(n)).^2 ) ),
     SNR_wy_dB = 10*log10( sum( s(n).^2 ) / sum( (s(n)-y(n)).^2 ) ),
     n=1:Nx-P;
     figure; subplot(111); plot(t(n),s(n),'k',t(n),d(n),'r',t(n),y(n),'b'); grid;
     xlabel('time [s]');
     title('Reference (BLACK), Noisy (RED), Filtered (BLUE)'); pause
  end
  figure; subplot(111); plot(1:M+1,h_opt,'ro-',1:M,h,'bx-'); grid; title('h(n): Wiener (RED), our (BLUE)');
  xlabel('n'); pause
