%% Testing Matlab functions for ONLY ONE-COMPONENT damped sine
close all, clear all, clc

signal = 3;       % 1 - low freq, very fast decay = very fast exp()
                  % 2 - very low freq, very slow decay = almost sin()
                  % 3 - mid freq, mid decay = typical, both exp() and sin()
                  % 4 - mechanical spectroscopy
N = 500;          % number of samples
SNR = 120;         % signal to noise ratio

%% Test signals
K = 1;            % number of components, in this program always 1
n = 0 : N-1;      % sample indexes
if( signal == 1 ) % low frequency, fast decay
    Om = 10.2*(2*pi)/N;     % 10.2
    D = 1e-1;               % 0.1
    A = 1.3; p = -pi/2;
end
if( signal == 2 ) % very low frequency, very slow decay
    Om = 5.2*(2*pi)/N;      % 5.2
    D = 1e-4;               % 0.0001
    A = 2.3; p = -pi/4;
end
if( signal == 3 ) % typical frequency, typical decay
    Om = 20.2*(2*pi)/N;      % 20.2
    D = 1e-2;                % 0.01
    A = 2.3; p = -pi/4;
end
    
x  = A * exp(-n*D) .* cos(Om*n+p);  % damped sin    
y = awgn( x, SNR );                 % Gaussian noise, SNR in [dB]

%% Plot signal
figure;
subplot(211); plot(x); grid; title('x(n)'); 
subplot(212); plot(y); grid; title('y(n) = x(n) + noise'); pause

%% Matlab functions

L = 2*K;  % [a(0) a(1) a(2) ] = [ 1, a1, a2 ]

% LPC
a1 = lpc(y,L);
p = roots(a1); p = log(p); De(1) = -real(p(1)); Ome(1) = imag(p(1));

% Levinson
r = xcorr(y,'biased');      % biased=1/N, unbiased=1/(N-k)
r = r(N:N+L);               % THE SAME AS #1
a2 = levinson(r,L);
p = roots(a2); p = log(p); De(2) = -real(p(1)); Ome(2) = imag(p(1));

% Yule-Walker autocorrelation method
a3 = aryule(y,L);
p = roots(a3); p = log(p); De(3) = -real(p(1)); Ome(3) = imag(p(1));

% Covariance
a4 = arcov(y,L);
p = roots(a4); p = log(p); De(4) = -real(p(1)); Ome(4) = imag(p(1));

% Burg
a5 = arburg(y,L);
p = roots(a5); p = log(p); De(5) = -real(p(1)); Ome(5) = imag(p(1));

% Prony
NB=1; NA=L;
[b,a6] = prony(y,NB, NA);
p = roots(a6); p = log(p); De(6) = -real(p(1)); Ome(6) = imag(p(1));

% Steiglitz-McBride
[b,a7] = stmcb(y,NB,NA);
p = roots(a7); p = log(p); De(7) = -real(p(1)); Ome(7) = imag(p(1));

C = 4; NI = 10; r=3;  % C=4, r=3 are the best
[Om8,  D8 ] = fHilbert(y,K);       De(8)  = D8;   Ome(8)  = Om8; 
[Om9,  D9 ] = fProny(y,K);         De(9)  = D9;   Ome(9)  = Om9;
[Om10, D10] = fSTMB(y,K,NI);       De(10) = D10;  Ome(10) = Om10;
[Om11, D11] = fKT(y,K);            De(11) = D11;  Ome(11) = Om11;
[Om12, D12] = fMatPen(y,K);        De(12) = D12;  Ome(12) = Om12;
[Om13, D13] = fYuleWalker(y,K,C);  De(13) = D13;  Ome(13) = Om13;
[Om14     ] = fPisarenko(y,K,C);   De(14) = Inf;  Ome(14) = Om14;
[Om15     ] = fESPRIT(y,K);        De(15) = Inf;  Ome(15) = Om15;
[Om16, D16] = fIpDFT_BY1_LC(y,NI); De(16) = D16;  Ome(16) = Om16;
[Om17, D17] = fFastTLS(y,r);       De(17) = D17;  Ome(17) = Om17;

methods = [ ...
'lpc        ';
'levinson   ';
'yule-walker';
'arcov      ';
'arburg     ';
'prony      ';
'stmb       ';
'fHilbert   ';
'fProny     ';
'fSTMB      ';
'fKT        ';
'fMatPen    ';
'fYuleWalker';
'fPisarenko ';
'fESPRIT    ';
'fIpDFT     ';
'fFastTLS   '];

fprintf('input      : Om  = %12.8f, D  = %12.8f \n', Om, D);
[dummy, ind] = sort( abs(Ome-Om), 'ascend');
disp('BEST OMEGA ========================================');
for k = 1 : length(De)
    fprintf('%s: Ome = %12.8f, De = %12.8f \n',methods(ind(k),:),Ome(ind(k)),De(ind(k)));
end
disp('MINIMUM ERROR OMEGA [%] ===========================');
for k = 1 : length(De)
    fprintf('%s: Ome = %12.8f, De = %12.8f \n',methods(ind(k),:),100*abs(Ome(ind(k))-Om)/Om,100*abs(De(ind(k))-D)/D);
end
[dummy,ind] = sort( abs(De-D), 'ascend');
pause
disp('BEST DAMPING ======================================');
for k = 1 : length(De)
    fprintf('%s: Ome = %12.8f, De = %f10 \n',methods(ind(k),:),Ome(ind(k)),De(ind(k)));
end
disp('MINIMUM ERROR DAMPING [%] =========================');
for k = 1 : length(De)
    fprintf('%s: Ome = %12.8f, De = %12.8f \n',methods(ind(k),:),100*abs(Ome(ind(k))-Om)/Om,100*abs(De(ind(k))-D)/D);
end