function [SMR] = MP2psycho(bx, fs, show)
%#############################################################################
%  Psycho-acoustical model for MP2 MEEG-audio layer
%  Input: bx - buffer with input signal, fs - sampling frequency
%  Ouput: SMR (Signal-to-Mask Ratio) 
%#############################################################################

% MP2
persistent PSYinit
persistent M BLKSIZE HBLKSIZE window CBlen 
persistent critband bminval widthn
persistent delay offs 
persistent r r1 r2 f f1 f2
persistent partition numlines cbval wlowb whighb sprdf minval tmn rnorm absthr
persistent nbb_1 nbb_2

%###################################################################################
% Initialization - only once, during first function call, then persistent constansts
%###################################################################################

if( isempty( PSYinit) ) 
 % General parameters  
   PSYinit = 1;              % initialization flag, set to 1 if INIT is done
   layer = 2;                % default layer
   M = 32;                   % number of filters in filter bank 
   BLKSIZE = 1024;           % FFT length - long data block
   HBLKSIZE = 513;           % half of the length
   delay = 480; offs = 576;  % 
 % Frequencies of critical bands 
   critband = [ 0 100 200 300 400 510 630 770 920 1080 1270 1480 1720 2000 2320 ...
                2700 3150 3700 4400 5300 6400 7700 9500 12000 15500 25000 30000 ];
 % Lower limit of signal from noise distance in each critical band, deciding about signal "unmasking"
   bminval = [ 20.0 20.0 20.0 20.0 20.0 17.0 15.0 10.0 7.0 4.4 4.5 4.5 4.5 4.5 4.5 ...
               4.5  4.5  4.5  4.5  4.5  4.5  4.5  4.5 4.5 3.5 3.5 3.5 ];
 % Indicator about acoustical width of M=32 filter subbands: 0=wide, 1=narrow
   widthn = [ 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ]; 
 % Long 1024-point Hannin window
   n=0:1:BLKSIZE-1; window = 0.5*(1-cos(2.0*pi*(n-0.5)/BLKSIZE));
 % Initialization of spectral buffers for: r = absolute value, f = phase
   temp = zeros(1,HBLKSIZE); r = temp; r1 = temp; f = temp; f1 = temp;
 
 % Initialization of tables with constants
 % fthr[ HBLKSIZE ]      = bark values corresponding to all FFT lags
 % partition[ HBLKSIZE ] = to what BARK-1/3 partition belongs each FFT coefficient 
 % wlowb[ CBlen ]        = first FFT lag in each BARK-1/3 partition
 % whighb[ CBlen ]       = last  FFT lag in each BARK-1/3 partition
 % numlines[ CBlen ]     = number of FFT lags in each BARK-1/3 partition (whighb-wlowb+1)
 % cbval[ CBlen ]        = central bark value in each BARK-1/3 partition
 % WARNING:              = count is counter of FFT lags in each BARK-1/3 partition

   dfreq = fs/BLKSIZE;  % step in frequency 
 
 % Calculate fthr, i.e. bark values corresponding to all FFT coefficients
   for k = 1:1:HBLKSIZE
       fk = (k-1)*dfreq; j = 2;
       while( fk > critband(j) ) j = j+1; end
       fthr(k) = (j-1) + (fk-critband(j-1))/(critband(j)-critband(j-1)) - 1;
   end

 % Calculate: partition, wlowb, whighb, numlines, cbval 
   partition(1) = 1; 
   count = 1; wlowb = 1; whighb = [];
   cbval(1) = fthr(1);
   bval_lo = fthr(1);
   for i = 2 : HBLKSIZE
      if((fthr(i)-bval_lo) > 0.33)
          partition(i) = partition(i-1) + 1;
          cbval( partition(i-1) ) = cbval( partition(i-1) ) / count;
          cbval( partition(i) ) = fthr(i);
          bval_lo = fthr(i);
          numlines( partition(i-1) ) = count;
          count = 1; wlowb = [ wlowb i]; whighb = [whighb i-1]; 
      else 
          partition(i) = partition(i-1);
          cbval( partition(i) ) =  cbval( partition(i) ) + fthr(i);
          count = count+1;
      end
   end
   numlines( partition(i-1) ) = count;                      % DFT lines in BARK-1/3
   cbval( partition(i-1) ) = cbval( partition(i-1) )/count; % central bark value in BARK-1/3
   whighb = [ whighb 513 ];
   CBlen = length(cbval);
  
 % Calculate "speading function"
   for(i=1:1:CBlen)
      for(j=1:1:CBlen)
         tempx = (cbval(i) - cbval(j))*1.05; % TZ: was bval insted of cbval
         if(tempx>=0.5 && tempx<=2.5)
            temp = tempx - 0.5;
            x = 8.0 * (temp*temp - 2.0 * temp);
         else x = 0.0;
         end
         tempx = tempx + 0.474;
         tempy = 15.811389 + 7.5*tempx - 17.5*sqrt(1.0+tempx*tempx);
         if (tempy <= -100.0) 
            sprdf(i,j) = 0.0;
         else
            sprdf(i,j) = exp( (x + tempy)*0.2302585093 );
         end
      end
   end
 % size(sprdf), pause; sprdf(:,1), pause

 % Calculate "minval", signal-to-noise ratio in each subband, ensuring unmasking
 % Calculate TMN (Tone Masking Noise)
 % Calculate normalization coefficients for "spreading" function
   for kb = 1 : CBlen 
       k = cbval(kb) + 0.5; minval(kb) = bminval(floor(k)+1);
       temp = 15.5 + cbval(kb);
       if( temp>24.5 ) tmn(kb) = temp; else tmn(kb) = 24.5; end
       rnorm(kb) = 0;
       for lb = 1 : CBlen
           rnorm(kb) = rnorm(kb) +  sprdf(kb,lb);
       end
   end 
 % tmn, pause; rnorm, pause

 % Absolute threshold of hearing, calculated from eq. (X.Y)
   freq = (0:HBLKSIZE-1)*dfreq/1000;
   absthr = 3.64*freq.^(-0.8) - 6.5*exp(-0.6*(freq-3.3).^2) + 10^(-3)*freq.^4; % in dB
   absthr = 10.^(absthr/20); % przelicz na poziom energii 
  
 % Summary:  
 % disp('nr wlowb whighb, nlines bval minval tmn');
   if(show==1) psycho = [ (1:CBlen)' wlowb' whighb' numlines' cbval' minval' tmn'], pause, end
 
end % end of PSYinit, start of main computations
  
%#########################################################################
% ANALYSIS
%#########################################################################

  k = 1:HBLKSIZE; k0 = 1:HBLKSIZE;

  for l = 1 : 2 % two times repeating for MP2   
   
    % Taking signal fragment
      bx = bx( 1+(l-1)*offs : length(bx) ); % plot(bx); pause
      
    % Shifting buffers
      r2 = r1; r1=r;
      f2 = f1; f1=f;
      
    % FFT of BLKSIZE=1024 point signal fragment
      X = fft( window .* bx(1:BLKSIZE) );
      r = abs(X(k)); f = angle(X(k));
 
    % Calculation of prediction coefficients for spectrum magnitude and phase
      pr = 2*r1 - r2;
      pf = 2*f1 - f2;
      
    % Calcualtion of spectral un-predictability measure  
      cw(k0) = sqrt( (r(k0).*cos(f(k0)) - pr(k0).*cos(pf(k0))).^2 + ...
               (r(k0).*sin(f(k0)) - pr(k0).*sin(pf(k0))).^2 ) ./ (r(k0) + abs(pr(k0)));
      
    % Calculation of energy "eb" and weighted un-predictability "cb" in BARK-1/3 partitions
      rr = r .* r;  rrcw = rr .* cw;
      for b = 1 : CBlen
          eb(b) = sum(   rr( wlowb(b) : whighb(b) ) );  
          cb(b) = sum( rrcw( wlowb(b) : whighb(b) ) );
      end 
       
    % Convolution with "spreading" function
      ecb = sprdf*eb';
      ctb = sprdf*cb';
      cbb = ctb' ./ ecb';
      enb = ecb' ./ rnorm;

    % Calculation of tonalty index "tbb" from "cbb"
      tbb = -0.299 - 0.43*log(cbb);
      
    % Calculation of SNR in each BARK-1/3 partition
      nmt = 5.5; work = tbb.*tmn + (ones(1,CBlen)-tbb)*nmt;
      snrb = max( minval(1:CBlen), work);  % TZ: minval?
 
    % Calculation of power coefficient    
      bcb = 10.^(-snrb/10);

    % Calculation of energy threshold  
      nbb = enb .* bcb;
      
    % Normalization by partition bandwidth
      nbw = zeros(1, HBLKSIZE);
      for kb = 1 : CBlen
          nbw( wlowb(kb) : whighb(kb) ) = nbb(kb) ./ (whighb(kb) - wlowb(kb)+1);
      end
          
    % Considering absolute threshold of hearing
      thrw = max(nbw, 10.^(absthr/10));
        
    % Calculating SMR for M=32-band filter MP2 filter bank - used for bit allocation
      for n = 1 : M  
          wlown = 1 + (n-1)*16; whighn = 1 + n*16;
          epartn(n) = sum( rr( wlown : whighn ) );  
          if( widthn(n)==1 )  npartn(n) = sum( thrw( wlown : whighn ) );
          else                npartn(n) = min( thrw( wlown : whighn ) ) * (whighn-wlown+1); end
      end
      SMR(l,1:M) = 10*log10( epartn ./ npartn + eps );  % output vector for MP2
      
   end  %  End of the analysis loop
   
% ##################################################################
  SMR = max( SMR );
  if(show==1) plot(SMR); xlabel('subband k'); title('SMR(k)'); pause, end  
% ##################################################################
