function [ns, IQs] = timing_recovery( IQ, alg, Nsps, n1st )
% Calculation of symbol positions (ns) and carrier states values (IQs)

  N = length(IQ); I = real(IQ); Q = (imag(IQ));
% Adaptation parameters
  damp = sqrt(2)/2;                                      % adaptation loop damping
  band = (0.5*pi/500) / (damp+1/(4*damp));               % adaptation loop bandwidth  
  mi1  = (4*damp*band) / (1 + 2*damp*band + band*band);  % adaptation coeff 1
  mi2  = (4*band*band) / (1 + 2*damp*band + band*band);  % adaptation coeff 2

% #################################################################
% Checking detection characteristics of all timing recovery methods
if(1) % select 0/1 = NO/YES
   N = floor(N/Nsps)*Nsps; mN = 1 : Nsps : N-4*Nsps+1; 
   for n=0:2*Nsps-1
      if(alg==1) % Gardner - frequency offset sensitive
         a = (I(n+Nsps+mN) - I(n+mN)) .* I(n+Nsps/2+mN); 
         cost(n+1) = mean(a);
      end
      if(alg==2) % Gardner - frequency offset sensitive
         a = (Q(n+Nsps+mN) - Q(n+mN)) .* Q(n+Nsps/2+mN); 
         cost(n+1) = mean(a);
      end
      if(alg==3) % Gardner - frequency offset sensitive
         a = (I(n+Nsps+mN) - I(n+mN)) .* I(n+Nsps/2+mN); 
         b = (Q(n+Nsps+mN) - Q(n+mN)) .* Q(n+Nsps/2+mN);
         cost(n+1) = mean(a + b);
      end
      if(alg==4) % Gardner - frequency offset insensitive
         cost(n+1) = mean(real((conj(IQ(n+Nsps+mN))-conj(IQ(n+mN)) .* IQ(n+Nsps/2+mN))));  
      end
      if(alg==5) % Mueller-Muller
         a = I(n+mN).*sign(I(n+Nsps+mN)) - I(n+Nsps+mN).*sign(I(n+mN));
         b = Q(n+mN).*sign(Q(n+Nsps+mN)) - Q(n+Nsps+mN).*sign(Q(n+mN));
         cost(n+1) = mean(a + b);
      end
   end
   figure; plot(0:2*Nsps-1,cost,'b.-'); xlabel('n'); ylabel('C(n)');
   title('Detection curve'); grid; pause
end
%##

% #######################################################
% Adaptive timing recovery for highly oversampled signals  
if( Nsps > 2 ) % big Nsps: choosing best sample
k=1; ns(1) = n1st; offs = 0; adap1 = 0; adap2 = 0;
for n = n1st : Nsps : length( IQ )-2*Nsps
   if(alg==1) % Gardner - frequency offset sensitive
      a = (I(n+Nsps+offs) - I(n+offs)) .* I(n+Nsps/2+offs); 
      err = -a;
   end
   if(alg==2) % Gardner - frequency offset sensitive
      a = (Q(n+Nsps+offs) - Q(n+offs)) .* Q(n+Nsps/2+offs); 
      err = -a;
   end
   if(alg==3) % Gardner - frequency offset sensitive
      a = (I(n+Nsps+offs) - I(n+offs)) .* I(n+(Nsps)/2+offs); 
      b = (Q(n+Nsps+offs) - Q(n+offs)) .* Q(n+(Nsps)/2+offs);
      err = -(a + b);
   end
   if(alg==4) % Gardner - frequency offset insensitive
      err = -real((conj(IQ(n+Nsps+offs)) - conj(IQ(n+offs))) .* IQ(n+(Nsps)/2+offs));  
   end
   if(alg==5) % Mueller-Muller
      a = I(n+offs).*sign(I(n+Nsps+offs)) - I(n+Nsps+offs).*sign(I(n+offs));
      b = Q(n+offs).*sign(Q(n+Nsps+offs)) - Q(n+Nsps+offs).*sign(Q(n+offs));
      err = -(a + b);
   end
   adap2 = adap2 + mi2 * err;                 % 1-nd update
   adap1 = adap1 + adap2 + mi1 * err;         % 2-st update
   while(adap1 >  1) adap1 = adap1 - 1; end   % # wrapping to interval (-1,1)
   while(adap1 < -1) adap1 = adap1 + 1; end   % # (mapping)
   offs = round( adap1 * Nsps );              % offset calculation as % of Nsps
        partition(k) = adap1;                 % for figure
        error(k)  = err;                      % for figure
        offset(k) = offs;                     % for figure
   k = k+1;                                   % index update 
   ns(k) = n+Nsps+offs;                       % storing symbol position
end
IQs = IQ(ns);                                 % estimated carrier states 
k = n1st : Nsps : length(IQ)-1;               % for figures
end % Nsps > 2
% ############

% #####################################################################
% Adaptive timing recovery for critical sampling - signal interpolation
if( Nsps == 1 || Nsps == 2 ) % Nsps = 1 or 2: Muller (1) or Gardner (2) method 
% Farrow filtration for Lagrange quadratic polynomial interpolation
  x2 = filter( [ 1/2  -1   1/2], 1, IQ );  x2 = x2(3:end);  % sample before
  x1 = filter( [ 1/2   0  -1/2], 1, IQ );  x1 = x1(3:end);  % sample central
  x0 = filter( [  0    1    0 ], 1, IQ );  x0 = x0(3:end);  % sample after
  adap1 = 0; adap2 = 0; k = 1; 
  for n = n1st+1 : Nsps : length(x0)-1
      xm1 = x2(n-1) * adap1*adap1 + x1(n-1) * adap1 + x0(n-1);
      xc0 = x2(n)   * adap1*adap1 + x1(n)   * adap1 + x0(n);
	  xp1 = x2(n+1) * adap1*adap1 + x1(n+1) * adap1 + x0(n+1);
      if(Nsps==1) % Mueller-Muller  
         a = real(xc0).*sign(real(xp1)) - real(xp1).*sign(real(xc0));
         b = imag(xc0).*sign(imag(xp1)) - imag(xp1).*sign(imag(xc0));
         err = -(a + b);
      end
      if(Nsps==2) % Gardner
         err = (real(xp1)-real(xm1))*real(xc0) + ...
               (imag(xp1)-imag(xm1))*imag(xc0);
      end
      adap2 = adap2 + mi2 * err;                 % 1-nd update
      adap1 = adap1 + adap2 + mi1 * err;         % 2-st update
      IQs(k) = xc0;                              % carrier state value
           partition(k) = adap1;                 % for figure
           error(k)  = err;                      % for figure
           offset(k) = 0;                        % for figure
      ns(k) = n+adap1;                           % storing symbol position
      k = k+1;                                   % index update 
  end
  k = n1st+1 : Nsps : length(IQ)-1;              % for figure
end % Gardner & Mueller-Muller
% ############################

    figure;
    subplot(311); plot(partition); grid; xlabel('n'); title('Partition(n)');
    subplot(312); plot(error); grid; xlabel('n'); title('Error(n)');
    subplot(313); plot(offset); grid; xlabel('n'); title('Offset(n)'); pause
    Ns=length(ns); nn = round(Ns/5) : Ns-round(Ns/5); 
    figure;
    subplot(221); plot(real(IQ(k)),imag(IQ(k)),'bo'); grid; xlabel('I(k)'); ylabel('Q(k)'); title('IN Q(k)=f(I(k))');
    subplot(223); plot(k,real(IQ(k)),'bo',k,imag(IQ(k)),'r*');  grid; xlabel('k'); title('IN I(k),Q(k)');
    subplot(222); plot(real(IQs(nn)),imag(IQs(nn)),'bo'); grid; xlabel('I(k)'); ylabel('Q(k)'); title('OUT Q(k)=f(I(k))');
    subplot(224); plot(ns,real(IQs),'bo',ns,imag(IQs),'r*');  grid; xlabel('k'); title('OUT I(k),Q(k)');
    pause
