% lab22_ex_adsl.m
% Simulation of multi-carrier transmission used in ADSL modems
clear all; close all;

  OCTAVE = 0;           % 0/1

% PARAMETERS
  Le = 16;              % length of the TEQ equalizer
  Lp = 32;              % length of the cyclic prefix (CP)
  N = 512;              % length of the DMT/OFDM frame, FFT size
  Np = N + Lp;          % length of the DMT/OFDM frame with cyclic prefix
  Niter = 300;          % number of transmitted frames (can be changed)
  fs = 2.208e6;         % sampling frequency [Hz]
  NoiseDbmHz = -160;    % power level of noise  [dBm/Hz]
  SignalDbm = 23;       % power level of input signal in the receiver [dBm]
  f0 = 100.5*fs/N;      % frequency of narrow-band disturbance
  A0 = 0;               % amplitude of narrow-band disturbance: 0 or 50
  CodingGain = 4.2;     % coding gain [dB]; e.g. without Viterbi enc/dec=0, with it=4.2
  CodingMargin = 6;     % SNR margin [dB]
  
  FilterTel = 1;        % 1 = high-pass filter, 0 = without it
  ADCppm = 5;           % ADC ppm error in receiver, e.g. -5, 0, 5
  
  CounterADC = 50;      % number of loops for ADC sampling error estimation
  CounterChan = 50;     % number of loops for impulse response estimation
  CounterSynchro = 50;  % number of loops for synchronization
  CounterSNR = 50;      % number of loops for SNR calculation and bit allocation

  TransmitDelay = 100;  % number of samples of the transmission delay
  NR1 = 20;  NR2 = 80;  % numbers of pilot carriers, may be used in the receiver
  NR3 = 60;  NR4 = 120; % numbers of data  carriers, observed in the receiver
  rand('state',0); randn('state',0);  % initialization of random generators
  SignalPower = 0.001 * 10^( SignalDbm/10 );            % to linear scale
  NoisePower   = 0.001 * fs/2 * 10^( NoiseDbmHz/10 );   % to linear scale

% Reading impulse response of the transmission line
  if(FilterTel == 0) h = load('DSLine.dat')';     % csaloop2 from [Arslan]
  else               h = load('DSLineHP.dat')';   % csaloop2 with high-pass filter
  end
% Generation of training sequence for channel impulse response identification 
  QAM4 = [ 1+j, -1+j, 1-j, -1-j ];        % 4-QAM constellation: 00, 01, 10, 11
  number = floor((4-eps)*rand(1,N/2))+1;  % random sequence of numbers { 1, 2, 3, 4 }
  X = QAM4( number );                                 % generation of the spectrum half
  X( 1 ) = sqrt(2); X( 257 ) = sqrt(2);               % DC and fs/2 are not used
  X( 258:512 ) = conj( fliplr( X(2:256) ) ); X = X.'; % conjugate symmetry
  x = real( sqrt(N/2)* ifft(X) ); st = x/std(x);      % transmitted time signal
  Swe = X;                                            % its spectrum
  Swy = zeros(N,1);                                   % accumulator for received spectra

% ####################################################################################
% MAIN LOOP - BEGINNING   ############################################################
% ####################################################################################

  Niter = CounterADC + CounterChan + CounterSynchro + CounterSNR; % all iterations
  nr = 2 : N/2;                    % numbers of used frequency carriers with data
  Nb = N;                          % DSL/OFDM frame length
  bx  = zeros(3*Nb,1); by = bx;    % input and output buffer for TX and RX signal
  bye = zeros(2*Nb,1);             % buffer for signal after TEQ equalizer
  S  = zeros(N,1); S1 = S;         % the last sent frame and one frame before
  sig = zeros(N/2-1,1); err = sig; % accumulators for signal and error energy
  scSUM = zeros(1,N+Lp);           % accumulator for the synchronization function
  QamHistory = [];                 % for constellation states for the carrier NR-th
  CopyCounterChan = CounterChan; TxDelay = TransmitDelay;       % set initial values
  ADCCorrect=1; ChanEstim=0; Synchronization=0; SNREstim=0;     % control parameters

  ppm = zeros(1,CounterChan); cfo = zeros(1,CounterChan); cnt = 0;
  n0 = 1;                          % first sample of narrow-band disturbance
  Nmax = Niter*Np;                 % maximum number of signal samples
  nnorm = 0:Nmax-1; n1n = 1;       % signal sample indexes without ADCppm 
  %Nppm=Nmax+(ADCppm*1e-6)*Nmax; nppm=0 :Nppm/Nmax: (Nmax-1)*Nppm/Nmax; n1p=1; % ADCppm
  Nppm=Nmax+(ADCppm*1e-6)*Nmax;  nppm=0 :Nmax/Nppm: (Nmax-1)*Nmax/Nppm; n1p=1; % ADCppm
  dfref = ADCppm*1e-6*fs,
  
  for iter = 1:Niter  % ------------------------------------------------------------
      
    % iter
      
    % Generation of data to be transmitted
      if( ADCCorrect==1 || ChanEstim==1) s = st; end % ADC & channel estimation
      if( Synchronization == 1 )                     % Synchronization
         s = sqrt(3)*(2*rand(N, 1)-1);               % uniform noise in [-1, 1]
      end
      if( SNREstim == 1 )                     % Data transmission:
      %  disp('###'); iter, pause 
         S0 = S1; S1 = S;                         % remember two last sent spectra
         number = floor((4-eps)*rand(1,N/2-1))+1; % random sequence of: 00, 01, 10, 11
         S = QAM4( number );                      % generation of the first spectrum half
         S(NR1-1) = 1+j; S(NR2-1) = -1-j;         % optional states of pilot signals
         S = [0 S 0 conj(fliplr(S))]; S =S.';     % if conjugate symmetry: ifft(S) is real!
         s = real( sqrt(N/2)* ifft(S) );          % inverse FFT, scaling
      end
      s = s * sqrt( SignalPower ) * sqrt( 10^( CodingGain/10 ) );
      
      %  Adding cyclic prefix
      if( ADCCorrect || ChanEstim == 1 ) x = s; % Channel estimation: without the prefix
      else x = [ (s(N-Lp+1:N)); s ];            % Synchronization and TX: add prefix
      end
      
      % Convolution with the original (un-shortened) channel impulse response
      bx = [ bx(Nb+1 : 3*Nb); x ];                           % put data into the buffer
      y  = conv( bx(Nb+1-TxDelay : 3*Nb- TxDelay ),h );      % convolution
      y = y(Nb+1 : Nb+Nb);                                   % cut the result

      % ADC sampling error in the receiver
      y(1:Nb,1)=interp1(nnorm(n1n : n1n+Nb-1),y(1:Nb,1),nppm(n1p : n1p+Nb-1),'spline');
      n1n = n1n + Nb; n1p = n1p + Nb;
      % For Octave ONLY - interp1() function returns NA for 1,N when ADCppm~=0
      if( OCTAVE==1  && ADCppm~=0 )
         if( isna(y(1)) )  y(1)  = y(2);    end
         if( isna(y(Nb)) ) y(Nb) = y(Nb-1); end
      end
      
      % Adding noise (dBm/Hz) and narrow-band disturbance 
      noise = sqrt( NoisePower ) * randn(Nb,1); 
      y = y + noise;
      nband = A0*sin(2*pi*f0/fs*(n0:n0+Nb-1)'); n0 = n0+Nb;
      y = y + nband;
      
      % Receiver: data scaling, taking into the buffer
      y = y / ( sqrt( SignalPower ) * sqrt( 10^( CodingGain/10 ) ) );
      by = [ by(Nb+1 : 3*Nb); y ];
      
      % TEQ - channel impulse response shortening 
      if( ADCCorrect == 0 && ChanEstim == 0 )                     % after channel estim
           ye  = conv(by(1+ReceiveDelay : 2*Nb+ ReceiveDelay),e); % conv with TEQ
           ye = ye(Nb+1 : Nb+Nb);                                 % cut the result
           bye = [ bye(Nb+1 : 2*Nb); ye ];                        % store in the buffer
      end
   
   %  ADC SAMPLING RATE ESTIMATION AND CORRECTION  ###########################
      if( (ADCCorrect == 1 ) && (iter > 4) )              % check condition
          CounterADC = CounterADC-1                       % decrement the counter
          Y1 = fft( by(1:N) ); Y2 = fft( by(N+1:2*N) );   % spectra of two consecutive ... 
          Y21 = Y2 .* conj(Y1);                           % symbols, should be the same 
          ang = fftshift( angle( Y21 ) );                 % check phase increase
          Ksize = 30; kf = (-Ksize : Ksize)';             % neighborhood around 0Hz
          angi = ang( (N/2+1)+kf );                       % angles around 0Hz
          coef = polyfit(kf, angi, 1); ppm1=coef(1)/(2*pi); cfo1=coef(2); % pause
          cnt = cnt+1; ppm(cnt)=ppm1; cfo(cnt)=cfo1;      % store ppm and cfo 
        % figure; plot(kf,angi,'b-',kf,coef(2)+coef(1)*kf,'r-'); title('Angle(n)'); pause
        % Y1=sum(by(1:N).*exp(-j*2*pi/N*(k/(1-ppm1)+cfo1)*(0:N-1)))/N; % correction 
          if(CounterADC==0)                                    % end of ADCCorrect
             Koffs = round(0.25*cnt);                          % number of skipped ppms    
             ppm=sort(ppm); ppm=-mean(ppm(Koffs:end-Koffs+1)), % ppm estimation
             cfo=sort(cfo); cfo=mean(cfo(Koffs:end-Koffs+1)),  % cfo estimation
             pause                                             % do hardware ppm correction
             ADCCorrect=0; ChanEstim=1; iterADCCorrect = iter; % next operation
          end
      end    
          
   %  ESTIMATION OF THE CHANNEL IMPULSE RESPONSE AND SYSTEM DELAY ############
      if( (ChanEstim == 1) && (iter > iterADCCorrect + 4 )) % check condition
          CounterChan = CounterChan-1                     % decrement the counter
          Swy = Swy + fft( by(1:N) ) /sqrt(N/2);          % accumulate the spectrum
          if(CounterChan==0)                              % end of ChanEstim
             hest = real( ifft( Swy ./ Swe ) )/CopyCounterChan; hest=hest'; % h=?
             [ hmax, nrmax] = max(abs(hest));             % find maximum
             K=15; n1st = nrmax-K; ReceiveDelay = n1st-1; %
             if(n1st<=0) n1st = n1st + N; end             %
             hest = [ hest(n1st:end) hest(1:n1st-1)];     % rotate circularly
             figure; plot(hest); title('hest(n)'); pause  % figure
             [ e ] = AdslTEQ(hest, Le, Lp, 0);            % design TEQ filter
             he = conv(e, hest);                          % shorten hest(n)
             figure; plot(he); title('he(n)'); pause      % figure
             he = he(1:N);                                % leave N samples
             HE = fft(he).';                              % FFT of he(n)
             figure; plot(fs/N*(0:N-1),20*log10(abs(HE))); title('|H(f)|'); pause
             Nb = Np; bx=zeros(3*Nb,1); by=zeros(3*Nb,1); bye=zeros(2*Nb,1); % set
             ChanEstim = 0; Synchronization = 1; iterChanEstim = iter;       % next
          end % end of CounterChan = = 0
      end % ChanEstim == 1

    % SYNCHRONIZATION #############################################################
      if( (Synchronization == 1) && (iter > iterChanEstim + 4) ) % condition
          CounterSynchro = CounterSynchro-1                     % decrement counter
          for n = 0 : N+Lp-1                                    % synchro function
              q = sum(bye(n+1:n+Lp).*bye(n+N+1:n+N+Lp)); 
              r = sum(bye(n+1:n+Lp).^2) + sum(bye(n+1+N:n+Lp+N).^2);
              sc(n+1)=q/r;
          end
          scSUM = scSUM + sc;                                % accum synchro function
          
          if (CounterSynchro == 0)                           % end of synchronization
             figure; plot(scSUM); title('SYNCHRO function'); % optional figure
             SynchDelay = find(scSUM == max(scSUM));         % find argument of max
             SynchDelay = SynchDelay + Lp/2, pause           % take prefix into account
             HEC = HE .* exp(j*2*pi/N *(0:N-1)'*(SynchDelay-Lp));  % correction
             FEQ = 1./HEC;                                         % FEQ
             Nb = Np; bx=zeros(3*Nb,1); by=zeros(3*Nb,1); bye=zeros(2*Nb,1); % set
             Synchronization = 0; SNREstim = 1; iterSynchronization = iter;  % next
          end % CounterSynchro = = 0
       end % Synchronization = = 1
       
%  SNR ESTIMATION ################################################################
       if( (SNREstim == 1) && (iter > iterSynchronization + 4) )   % condition
         r = bye(1+SynchDelay : N+SynchDelay);   % synchronize, remove prefix, received
         R = sqrt(2/N) * fft(r);                 % FFT
         R = R .* FEQ;                           % FEQ correction
         
       % figure;
       % subplot(211); plot(nr,real(R(nr)),'ro',nr,real(S0(nr)),'bx'); grid;
       % subplot(212); plot(nr,imag(R(nr)),'ro',nr,imag(S0(nr)),'bx'); grid; pause
         SRerr = S0 - R;                         % reconstruction error
         sig = sig + S0(nr).*conj(S0(nr));       % accumulate signal energy in sub-bands
         err = err + SRerr(nr).*conj(SRerr(nr)); % accumulate error energy in sub-bands
         QamHistory = [ QamHistory; R(NR1) R(NR2) R(NR3) R(NR4) ];% NR-th sub-carriers
         
       end % SNREstim == 1
     
   end % main loop -------------------------------------------------------------
% ####################################################################################
% MAIN LOOP  END  ###################################################################
% ####################################################################################

% Figures: SNR in sub-bands, received values for the NR-th carrier (4-QAM is sent)
  SNR = (sig + eps) ./ (err + eps); SNRdB = 10*log10( SNR );
  figure; plot(nr, SNRdB,'b'); title('SNR [dB]'); xlabel('Channel number'); pause
  figure;
  subplot(221); plot( real(QamHistory(:,1)), imag(QamHistory(:,1)),'bx'); title('4QAM-No1');
  subplot(222); plot( real(QamHistory(:,1)), imag(QamHistory(:,2)),'bx'); title('4QAM-No2');
  subplot(223); plot( real(QamHistory(:,3)), imag(QamHistory(:,3)),'bx'); title('4QAM-No3');
  subplot(224); plot( real(QamHistory(:,4)), imag(QamHistory(:,4)),'bx'); title('4QAM-No4');
  pause
 
% Do yourself detection of constellation states (numbers), i.e. do the de-QAM
% ........................................................................

% Figure: ALLOCATION OF BITS for sub-carriers
  gamma = 10 ^ ( (9.8 + CodingMargin - CodingGain)/10 );
  ab = log2( SNR/gamma+1 ); ab = floor( ab );       % log2( snr/gamma +1 )
  idx = find( ab > 15 ); ab( idx ) = 15;            % no more than 15 bits per channel
  idx = find( ab < 2 ); ab( idx ) = 0;              % no less than 2 bits per channel
  figure; plot(nr, ab,'b'); xlabel('channel number'); ylabel('number of bits'); pause
  BitsPerSymbol = sum( ab ), BitsPerSecond = BitsPerSymbol * fs/Np, pause
  