% lab23_ex_dab_fic.m

% Simple Matlab program for decoding service information from one DAB frame
% Masterthesis
% Zurcher Hochschule fur Angewandte Wissenschaften
% Zentrum fur Signalverarbeitung und Nachrichtentechnik
% (c) Michael Häner-Höin, 12.4.2011 ZSN, info.zsn@zhaw.ch
% Extended:
% AGH University of Science and Technology, Krakow, Poland
% (c) Tomasz Zielinski, 31.12.2019, tzielin@agh.edu.pl

  clear all; close all; fclose('all');

% --------------------------
% PROGRAM CONTROL PARAMETERS
% --------------------------  

  CorrectADC_ON         = 1; % 0/1 CORRECT ADC sampling frequency by re-sampling 
  CorrectFreqOffset_ON  = 1; % 0/1 CORRECT carrier frequency offset (CFO)
  CorrectTimeSynchro_ON = 1; % 0/1 CORRECT time synchro using known DF (delta freq)
  Disturbances_ON       = 0; % 0/1 ADD disturbances
  
% --------------------------
% GENERAL DAB PARAMETERS
% --------------------------  

  fs = 2.048e6;            % sampling frequency used in the DAB algorithm

  NSymbPerFrame = 76;      % L, number of OFDM symbols per one DAB frame
  NCarrPerSymb  = 1536;    % K, number of transmitted carriers
  NSampPerFrame = 196608;  % Tf, duration of the transmission frame (in samples)
  NSampPerNull  = 2656;    % TNULL, duration of the Null symbol (in samples)
  NSampPerSymb  = 2552;    % Ts, duration of OFDM symbols with indices: l=1...L
  Nfft          = 2048;    % Tu, inverse of the carrier spacing
  NSampPerPrefix = 504;    % delta, duration of the guard interval

  NFIBsPerFrame = 12;      % (FIC) number of FIBs per one DAB frame (non-interleaved)
  NCIFsPerFrame =  4;      % (MSC) number of CIFs per one DAB frame (time-interleaved)
  NSymbPerFIC   =  3;      % (FIC) number of OFDM frames per one FIC
  NFIBsPerCIF   =  3;      % (FIC) number of FIBs per one CIF

  NSampPerData = NSymbPerFrame*NSampPerSymb; % DAB frame length without NULL
  
% Generation of Phase Reference Symbol (used for frame synchro) (ETSI pp. 147-149)
  [specPhaseRef, sigPhaseRef] = PhaseRefGen( 0 );
% Generation of a Frequency Interleaving and De-interleaving Tables (ETSI pp. 157-161)
  [FreqInterleaverTab]   = FreqInterleaverTabGen( Nfft );                       
  [FreqDeinterleaverTab] = FreqDeinterleaverTabGen( FreqInterleaverTab, Nfft );
% Generation of trellis for polynomial (ETSI, page 129-130, below fig. 72)
  VitTrellis = poly2trellis(7,[133 171 145 133]); % 7 = delay, then octal numbers
% Generation of matrices for OUR Viterbi decoder (SLOW!), working in Matlab and Octave
% [ Generator, Transition, Ref1, Ref2 ] = DABViterbiTZinit();
  [ Generator, InStates1, InStates2, RefOutValues1, RefOutValues2 ] = DABViterbiTZinit();

% Read an IQ file for FIC decoding
% DAB+ recordings: https://sdr.kt.agh.edu.pl/sdrdab-decoder/download.html
  Nframes = 3;                              % number of read DAB frames    
  if(1) % synthesized, disturbance-free
     ReadFile = fopen('DAB_PolishRadio_TX_IQ.dat', 'rb');  
     x = fread( ReadFile, [2, Nframes * NSampPerFrame], 'float' ); % from-to
     x = x(1,:) + j*x(2,:); x = x.' / 2^15;                        % complex IQ 
     fclose( ReadFile );
  else  % recorded, with disturbances
     ReadFile = 'DAB_PolishRadio_RX_IQ.wav';                % file name
     n1st = 1; nlast = n1st + Nframes * NSampPerFrame-1;    % from-to
     [x, fs] = audioread( ReadFile, [n1st, nlast] );        % read
     x = x(:,1) - j*x(:,2);                                 % complex IQ
  end
  figure; plot(real(x)); xlabel('n'); title('Real(x(n))'); grid; pause

% #########################################################################  
% Optional addition of disturbances
if( Disturbances_ON ) x = disturbDAB( x ); end

% #########################################################################
% Detection of first two NULL symbols - sum( abs( samples(...) ))
  xabs = abs(x);
  NSampPerSynchro = NSampPerFrame + NSampPerNull;
  
% First DAB frame  
  acc = zeros(1,NSampPerSynchro); acc(1) = sum( xabs(1:NSampPerNull) );
  for n = 1 : NSampPerSynchro; 
      acc(n+1) = acc(n) - xabs(n) + xabs(n+NSampPerNull);
  end
  [~, imin] = min(acc);
  i1st = imin, % first sample of DAB #1 frame w/ NULL
  
% Second DAB frame
  ADCppmMAX = 400;                           % assumed maximum ADC ppm error
  M = round( 0.5 * ADCppmMAX * NSampPerFrame/fs ); % maximum possible offset
  i2nd = imin + NSampPerFrame - M;           % first sample index to test 
  acc = zeros(1,2*M); acc(1) = sum( xabs( i2nd : i2nd + NSampPerNull-1 ) ); 
  for n = 1 : 2*M-1 
      acc(n+1) = acc(n) - xabs(i2nd+n) + xabs(i2nd+n+NSampPerNull);
  end
  [dummy imin] = min(acc); clear xabs;
  i2nd = i2nd + imin,  % first sample of DAB #2 frame /w NULL
  
% ADC correction of the first DAB frame starting from NULL symbol
  if( CorrectADC_ON )
     NSampPerFrameADC = i2nd - i1st,                % found frame length
     if( NSampPerFrameADC ~= NSampPerFrame )        % comparison with expected
         step = NSampPerFrameADC / NSampPerFrame;   % signal re-sampling
         x = interp1( [0:1:NSampPerFrameADC+NSampPerNull-1]', ...
             x(i1st : i1st+NSampPerFrameADC+NSampPerNull-1), ...
             [0:step:NSampPerFrameADC+NSampPerNull-1]','spline');
     end
  else
     x = x( i1st : i2nd -1 ); 
  end
  
% #########################################################################  
% Carrier frequency offset estimation using the PHASE REFERENCE symbol
% #########################################################################  

  i1st = 1 + NSampPerNull;    % skip NULL symbol

if( CorrectFreqOffset_ON == 0 )

  df = 0;

else

% ESTIMATION of CFO fractional part: df_fract 
  Ncut = 25;             % Assumed max time offset error +/-
  icp   = (i1st+Ncut      : i1st+NSampPerPrefix-Ncut-1)';       % CP indexes 
  isymb = (i1st+Ncut+Nfft : i1st+NSampPerPrefix-Ncut-1+Nfft)';  % SYMBOL indexes
  z =  sum( conj(x( icp )) .* x( isymb ) );    % correlation using cyclic prefix
  df_fract = angle(z)/(2*pi),
% df_fract COMPENSATION of the PHASE REFERENCE symbol
  icpsymb  = (i1st : i1st + NSampPerPrefix+Nfft-1)';            % CP+SYMBOL indexes 
  s = x(icpsymb) .* exp(-j*2*pi/Nfft*df_fract*(icpsymb));       % CFO correction

% ESTIMATION of CFO integer part: df_int
% Multiplication in time domain = cyclic convolution (corr) in frequency domain
% Cyclic correlation of signal spectrum with the PhaseRef spectrum --> find maximum
  M = 20;   % M = 10? 20?  max search region = [-M, +M] around 0 spectrum shift 
  isymb = (NSampPerPrefix+1 : NSampPerPrefix+Nfft)';
  XC = fft( s( isymb ) .* conj( sigPhaseRef( isymb ) ) );      % fft( mult in time)
  XC = [ XC(end-M+1:end); XC(1:1+M) ];
  [ val, indx ] = max( XC ); 
  df_int = indx - (M+1), % pause

% SUMMATION   
  df = df_int + df_fract;

  if( CorrectTimeSynchro_ON ) 
     % START: Extra time synchronization using PHASE REFERENCE SYMBOL
     % Frequency offset correction of the Phase Reference Symbol
       s = x( icpsymb ) .* exp(-j*2*pi/Nfft*df*( icpsymb ));
     % TIME OFFSET CALCULATION IN FREQ DOMAIN - mult in freq -> conv/corr in time  
       S = fft( s( isymb ) );
       work = ifft( S .* conj( specPhaseRef ) ) / NSampPerSymb; % ifft(mult in freq)
       M=7; work=[ work(end-M+1:end); work(1:M+1) ]; % Assumed max +/- 7
       [ dummy indx ] = max(abs(work(1:2*M+1))); 
       ioffsA = indx-(M+1),
     % TIME OFFSET CALCULATION IN TIME DOMAIN - direct signal correlation in time  
       M = 7; XC = []; Nsymb = NSampPerSymb;  % Assumed max +/-7
       XC(M+1) = sum( s .* conj(sigPhaseRef) )/Nsymb;
       for m = 1 : M
           XC( M+1+m ) = sum( s(1+m:Nsymb) .* conj(sigPhaseRef(1:Nsymb-m)) )/(Nsymb-m);
       end
       for m = 1 : M
           XC( M+1-m ) = sum( s(1:Nsymb-m) .* conj(sigPhaseRef(1+m:Nsymb)) )/(Nsymb-m);
       end
      [ dummy ioffsB ] = max(abs(XC));
       ioffsB = ioffsB - (M+1), pause
     % CORRECTION of DAB frame beginning (without NULL)
       ioffs = round( (ioffsA + ioffsB)/2 );               % use mean value 
       i1st = i1st + ioffs;
  end

end

% FINAL CFO CORRECTION OF THE WHOLE LONG DAB FRAME WITHOUT NULL WITH PHASEREF
  x = x(i1st : i1st + NSampPerData-1) .* exp(-j*2*pi/Nfft*df*(0 : NSampPerData-1)).' ;
  
% FFT
  Ncp = NSampPerPrefix; Ncar = NCarrPerSymb;  % shorter names
  offs = 0;                                   % take a few last CP samples
  X = zeros( [NSymbPerFrame, Ncar] );         % creating empty array for all FFTs
  for nsymb = 1 : NSymbPerFrame               % processing all OFDM symbols
    i = (nsymb-1) * NSampPerSymb;             % index of the first sample
    s = x( 1+i : Ncp+Nfft+i );                % samples of one whole OFDM symbol
    s = s( Ncp+1-offs : Ncp+Nfft-offs );      % take only Nfft part minus offs
    S = fft( s );                             % FFT of the selected samples
    S = fftshift( S );                        % negative frequencies to the left
    X( nsymb, 1 : Ncar/2 )      = S( Nfft/2-Ncar/2+1 : Nfft/2 );    % store to FFT array
    X( nsymb, Ncar/2+1 : Ncar ) = S( Nfft/2+2 : Nfft/2+Ncar/2+1 );  % used carriers only
  end
  DX = X( 2:NSymbPerFrame, : ) .* conj( X( 1:NSymbPerFrame-1, :) ); % phase shift
  
  figure; plot(angle(DX(:,:)),'b*'); xlabel('symbol k'); title('Phase Shifts'); grid; pause
  figure; plot(real(DX(:,100)),imag(DX(:,100)),'b*'); xlabel('I'); title('Q(k)=f(I(k))'); grid; pause

% Frequency de-interleaver - using known FreqDeinterleaverTab table
  DeintDX = zeros( size( DX ) );
  for r = 1 : length( FreqDeinterleaverTab(:,4) )                               
      DeintDX( :, FreqDeinterleaverTab(r,4) ) = DX(:,r);
  end
  
% D-QPSK demodulation: complex --> real, imag
% D-QPSK symbol de-mapper - from 2D matrix to 1D vector, ETSI pp. 157 &&&&&&
  DataFIC = zeros(1, NSymbPerFIC*NCarrPerSymb*2);                  % 3*1536*2
  for f = 1 : NSymbPerFIC % FIC DATA from OFDM symbols, first real, then imag 
      DataFIC( 1+(f-1)*NCarrPerSymb*2 : (f)*NCarrPerSymb*2 ) = ...
                                 [ real(DeintDX(f,:)) imag(DeintDX(f,:)) ];
  end

% Constants for FIC
  NSampAfterTimeDepunct = 3096; NSampAfterVit = NSampAfterTimeDepunct/4-6; % 768; 
 
% Put bits of FIBs corresponding to one CIF in separate row -------------------------------
  NficbitsPer1CIF = NSymbPerFIC*NCarrPerSymb*2/NCIFsPerFrame; % 2304/2304/3072/2304
  FIC = zeros(NCIFsPerFrame, NficbitsPer1CIF);       
  for r = 0:NCIFsPerFrame-1
      FIC(r+1,:) = DataFIC( r*NficbitsPer1CIF+1 : (r+1)*NficbitsPer1CIF );
  end
        
% De-puncturing using special function ------------------------------------
  DataDep = [ depuncturing( FIC(:,1:2016),16 ) depuncturing( FIC(:,2017:2292),15 ) ...
              depuncturing( FIC(:,2293:2304),8 )];
        
 % Viterbi decoder using Matlab function vitdec() -------------------------
   DataVit = zeros(NCIFsPerFrame,NFIBsPerCIF*256+6);  % Bitvector + Tail
   for f=1:NCIFsPerFrame                              % Matlab. +/-real or {-1,+1}
     % DataVit(f,:) = vitdec( DataDep(f,:), VitTrellis, 1, 'trunc', 'unquant'); % Matlab fast
       DataVit(f,:) = DABViterbiTZ(DataDep(f,:),Generator,InStates1,InStates2,RefOutValues1,RefOutValues2,0); % TZ, Octave slow
    end
    DataVit = DataVit(:,1:end-6);                     % Tail removing

% Energy dispersal using special function EnergyDispGen() -----------------
  DataEnerg = zeros(size(DataVit));
  for m=1:NCIFsPerFrame
      DataEnerg(m,:) = xor( DataVit(m,:), EnergyDispGen( NFIBsPerCIF*256 ) );
  end

% FIBs building -----------------------------------------------------------
  FIB = reshape( DataEnerg', 256,NFIBsPerFrame )';
        
% CRC checking of FIBS ----------------------------------------------------
  DABFrameNr = 1;                          % only one frame is analyzed
  FIBCRCCheck = zeros(1,NFIBsPerFrame);
  for k = 1 : NFIBsPerFrame
      FIBCRCCheck(k) = CRC16( FIB(k,:) );  % special function CRC16()
  end    
  if sum(FIBCRCCheck) == NFIBsPerFrame     % 12?
     disp('CRC OK');                       % FIGs building from FIBS 
     for FIBNr = 1  : NFIBsPerFrame        % to 12
         pos = 1;
         while pos < 241
               if FIB(FIBNr,pos:pos+7) == [1 1 1 1 1 1 1 1], break, end
             % FIG type and length finding
               Type = BinToDec( FIB(FIBNr,pos:pos+2), 3 );
               FIGDataLength = BinToDec( FIB(FIBNr,pos+3:pos+7), 5 ); % in bytes
             % FIG building (reconstruction) using special function FIGType()
               FIGType( DABFrameNr, FIBNr, Type, FIGDataLength, ...
                        FIB(FIBNr, pos+8:pos+8+8*FIGDataLength-1) )
               pos = pos + (1+FIGDataLength)*8;
         end
     end
     disp('################################## Ready !');
   % disp('PRESS ANY KEY !'); % pause
  else
     disp(['CRC Fail! Frame: ',num2str( DABFrameNr )]);
  end
